(* Librairie de calcul d'entiers positifs de longueur arbitraire *)

(******************************************************************************)
(* Les entiers de longueur arbitraire seront représentés par une liste        *)
(* d'entiers. Le choix de la représentation par liste plutôt que par vecteur  *)
(* permet de pas avoir à déterminer à l'avance le nombre de chiffre du        *)
(* résultat d'une opération. Elle pose des problèmes pour l'algorithme de     *)
(* comparaison qui demande un parcours partant de la fin, mais on peut adapter*)
(* l'algorithme. La découpe lors de l'algorithme diviser pour régné est moins *)
(* naturelle, mais on n'a pas de réelle perte de complexité.                  *)
(* L'utilisation de vecteur aurait pu permettre une implementation en place de*)
(* l'algorithme de karatsuba, mais qui est trop complexe pour être raisonnable*)
(* dans ce sujet.                                                             *)
(* Les algorithmes s'écrivent naturellement de manière récursive.             *)
(* Les unités seront placées en tête de liste, ce qui facilitera la gestion   *)
(* des retenues dans le calcul de l'addition .                                *)
(******************************************************************************)

(* Un entier de longueur arbitraire est une liste d'entier de type int *)
(* entre 0 et 9 inclus. Les derniers entiers pouvant eventuellements êtres *)
(* nuls *)
type entier = int list ;;

                               (* Conversions *)

(* Conversion d'un int en entier *)
let rec int_vers_entier ( n : int ) : entier =
  if n < 10 then [ n ]
  else ( n mod 10 ) :: ( int_vers_entier ( n / 10 ) ) ;;

(* Conversion d'un entier vers un int *)
let rec entier_vers_int ( n : entier ) : int = match n with
  | [] -> 0
  | chiffre :: nombre -> chiffre + 10 * ( entier_vers_int nombre ) ;;

                               (* Comparaisons *)

type ordre = PLUS_PETIT | EGAL | PLUS_GRAND ;;

(******************************************************************************)
(* -- Comparaison de deux entiers --                                          *)
(* On élimine les cas où les opérandes ne sont pas de la même longueur.       *)
(* Le résultat de la comparaison de deux nombres positifs de même longueur,   *)
(* consiste en le résultat de la comparaison des deux premiers chiffres       *)
(* différent des deux nombres en partant de la gauche dans la notation        *)
(* positionnelle des deux nombres.                                            *)
(* Ici la liste commence par le chiffre des unités. Pour éviter de renverser  *)
(* la liste avant de chercher le premier chiffre différent. On parcours les   *)
(* chiffres en partant de celui des unités et on sauvegarde à chaque appel    *)
(* récursif le résultat de la comparaison des deux derniers chiffres          *)
(* différents rencontrés.                                                     *)
(* Complexité : O(log(n)) (n est le plus grand des opérande)                  *)
(******************************************************************************)

let comparaison ( operande_g : entier  ) ( operande_d : entier ) : ordre =
      (* Résultat de la comparaison de deux int. *)
      let comparaison_int ( n : int ) ( m : int ) =
        if n < m then PLUS_PETIT
        else if n > m then PLUS_GRAND
        else EGAL in
      (* Parcours des chiffres des opérandes *)
      let rec parcour_chiffres_operande
          ( operande_g : entier )
          ( operande_d : entier )
          ( derniere_comparaison : ordre )
            (* la comparaison des derniers chiffres différents rencontrés *)
        : ordre = match operande_g, operande_d with
        | [], [] -> derniere_comparaison
        | [], chiffre :: suite when chiffre != 0 -> PLUS_PETIT
        | chiffre :: suite, [] when chiffre != 0 -> PLUS_GRAND
        | [], chiffre :: suite ->
            parcour_chiffres_operande [] suite derniere_comparaison
        | chiffre :: suite, [] ->
            parcour_chiffres_operande suite [] derniere_comparaison
        | chiffre_g :: suite_g, chiffre_d :: suite_d    ->
            let comparaison_chiffres = (comparaison_int chiffre_g chiffre_d) in
            let nouvelle_comparaison = match comparaison_chiffres with
              | EGAL -> derniere_comparaison
              | _ -> comparaison_chiffres in
            parcour_chiffres_operande suite_g suite_d nouvelle_comparaison in
      parcour_chiffres_operande operande_g operande_d EGAL ;;

let est_plus_petit ( operande_g : entier ) ( operande_d : entier ) : bool =
  comparaison operande_g operande_d == PLUS_PETIT ;;

let est_plus_grand ( operande_g : entier ) ( operande_d : entier ) : bool =
  comparaison operande_g operande_d == PLUS_GRAND ;;

let est_egal ( operande_g : entier ) ( operande_d : entier ) : bool =
  comparaison operande_g operande_d == EGAL ;;

                               (* Arithmétique *)

(******************************************************************************)
(* -- Addition --                                                             *)
(* On ajoute les chiffres un par un depuis le chiffre des unités à chaque     *)
(* appel récursif. On propage la retenu en la passant en paramètre au         *)
(* prochain appel récursif.                                                   *)
(* Complexité : 0(log(n))                                                     *)
(******************************************************************************)
let addition ( operande_g : entier ) ( operande_d : entier ) : entier =
  let rec addition_et_propagation
      ( operande_g : entier )
      ( operande_d : entier )
      ( retenue : int )
    : entier = match operande_g, operande_d with
    | [], [] -> if retenue == 0 then [] else [ retenue ]
    | [], _ -> addition_et_propagation [ 0 ] operande_d retenue
    | _, [] -> addition_et_propagation operande_g [ 0 ] retenue
    | chiffre_g :: suite_g, chiffre_d :: suite_d    ->
        let somme_chiffres = chiffre_g + chiffre_d + retenue in
        let chiffre_suivant = somme_chiffres mod 10 in
        let retenue_suivante = somme_chiffres / 10 in
        chiffre_suivant :: ( addition_et_propagation suite_g suite_d
                               retenue_suivante) in
  addition_et_propagation operande_g operande_d 0 ;;

(******************************************************************************)
(* -- Soustraction --                                                         *)
(* Après avoir réordonnés les opérandes dans l'ordre croissant, on soustrait  *)
(* les chiffres un par un depuis le chiffre des unités à chaque appel         *)
(* récursif. On ajoute les retenues et les propagent à chaque appel si        *)
(* besoin. On gère le cas où des 0 apparaissent en tête des nombres. Cas qui  *)
(* peux intervenir dans l'algorithme de multiplication.                       *)
(* On fait attention à ce que la taille des nombre soit préservée pour        *)
(* l'algorithme de multiplication                                             *)
(* Complexité : 0(log(n))                                                     *)
(******************************************************************************)
let soustraction ( operande_g : entier ) ( operande_d : entier ) : entier =
  let rec soustraction_bon_ordre
      ( operande_g : entier )
      ( operande_d : entier )
      ( retenue : int )
    : entier = match operande_g, operande_d with
    | [], [] -> []
    | chiffre_g :: suite_g, [] -> ( chiffre_g - retenue ) :: suite_g
    | chiffre_g :: suite_g, chiffre_d :: suite_d ->
        let retenue_suivante = if chiffre_g < ( chiffre_d + retenue )
          then 1 else 0 in
        let chiffre_suivant = chiffre_g - ( chiffre_d + retenue )
                              + 10 * retenue_suivante in
        chiffre_suivant::
          ( soustraction_bon_ordre suite_g suite_d retenue_suivante )
    | [], _ -> [] in
  let operande_plus_petit = if est_plus_petit operande_g operande_d
      then operande_g else operande_d in
  let operande_plus_grand = if est_plus_grand operande_g operande_d
      then operande_g else operande_d in
  soustraction_bon_ordre operande_plus_grand operande_plus_petit 0 ;;

(******************************************************************************)
(* -- Multiplication --                                                       *)
(* On utilisera l'algorithme de Karastuba.                                    *)
(* On implémente directement la multiplication pour des nombres de longueur   *)
(* arbitraire                                                                 *)
(******************************************************************************)

(* Coupe un entier de longueur pair en deux partie de longueur égales *)
(* Complexité : O(log(n)) *)
let couper_en_deux ( n : entier ) : entier * entier =
  let demi_taille = ( List.length n ) / 2 in
  let rec repartition
      ( n : entier )
      ( indice : int )
    : entier * entier =
    match n with
    | [] -> ( [], [] )
    | chiffre :: suite when indice < demi_taille ->
        let resultat_g, resultat_d = repartition suite ( indice + 1 ) in
        ( chiffre :: resultat_g, resultat_d )
    | chiffre :: suite ->
        let resultat_g, resultat_d = repartition suite ( indice + 1 ) in
        ( resultat_g, chiffre :: resultat_d ) in
  repartition n 0 ;;

(* La multiplication par une puissance de 10 est un simple décalage *)
(* Complexité : O(p) avec p la puissance de 10  *)
let rec multiplication_puissance_de_10 ( n : entier ) ( puissance : int )
  : entier = if puissance == 0 then n
      else multiplication_puissance_de_10 ( 0 :: n ) ( puissance - 1 ) ;;

(* Des entiers signés apparaissent dans le calcul *)
type signe = POSITIF | NEGATIF

(* Pour multiplier deux nombres de tailles différentes on complète par des 0  *)
(* Complexité : O(log(n)) *)
let rec equilibrer operande_g operande_d =
  match operande_g, operande_d with
  | [],[] -> [], []
  | chiffre_g :: suite_g, [] ->
      let suite_eq_g, suite_eq_d = equilibrer suite_g [] in
      chiffre_g :: suite_eq_g, 0 :: suite_eq_d
  | [], chiffre_d :: suite_d  ->
      let suite_eq_g, suite_eq_d = equilibrer [] suite_d in
      0 :: suite_eq_g, chiffre_d :: suite_eq_d
  | chiffre_g :: suite_g, chiffre_d :: suite_d ->
      let suite_eq_g, suite_eq_d = equilibrer suite_g suite_d in
      chiffre_g :: suite_eq_g, chiffre_d :: suite_eq_d ;;

(* On implémente l'algorithme de Karatsuba *)
(* Analyse de complexité dans le cas de deux nombres de l = 2^k chiffres : *)
(* Le cas de base est une multiplication entre deux entier en O(1) *)
(* Un appel récursif consiste en : *)
(* - Un équilibrage : O(l) *)
(* - Deux découpes : O(l) *)
(* - Deux soustraction sur des éléments de longueur l / 2 : O(l)  *)
(* - Trois appels récursif sur des éléments de taille l / 2  *)
(* - Deux comparaisons sur des éléments de taille l / 2 : O(l) *)
(* - Quatres additions / soustraction sur des éléments de taille l : O(l) *)
(* - Deux multiplication par 10 ^ m avec m = l ou m = l / 2 : O(l) *)
(* *)
(* Si T(l) est la complexité de l'algorithme en fonction de la longueur des *)
(* nombres. On a donc : T(l) = 3T(l/2) + O(l) . D'après le théorème maître on *)
(* a donc T(l) = O(l^(log_2(3))) ~= O(l^1.585) *)

let rec multiplication ( operande_g : entier ) ( operande_d : entier )
  : entier =
  let operande_eq_g, operande_eq_d = equilibrer operande_g operande_d in
  match operande_eq_g, operande_eq_d with
  | chiffre_g :: [], chiffre_d :: [] ->
      int_vers_entier ( chiffre_g * chiffre_d )
  | _ -> (
      let longueur = List.length operande_eq_g in
      let x0, x1 = couper_en_deux operande_eq_g in
      let y0, y1 = couper_en_deux operande_eq_d in
      let alpha = multiplication x1 y1 in
      let beta = multiplication x0 y0 in
      let gamma =
        multiplication ( soustraction x1 x0 ) ( soustraction y1 y0 ) in
      let signe_gamma =
        if ( est_plus_grand x1 x0 ) == ( est_plus_grand y1 y0 )
        then POSITIF else NEGATIF in
      let alpha_plus_beta = addition alpha beta in
      let sigma = match signe_gamma with
        | POSITIF -> soustraction alpha_plus_beta gamma
        | NEGATIF -> addition alpha_plus_beta gamma in
      addition
        ( addition
          ( multiplication_puissance_de_10 alpha ( ( longueur / 2 ) * 2 ) )
          ( multiplication_puissance_de_10 sigma ( longueur / 2 ) ) )
        beta
    ) ;;

(* On n'utilise pas en pratique la formule avec les additions car la retenue *)
(* peut modifier la taille des sommes et introduire plus de calcul  *)

(* Exponentiation rapide *)

let rec pow_entier n i =
  if i == 0 then [1]
  else if i mod 2 == 0 then pow_entier (multiplication n n) (i/2)
  else multiplication n ( pow_entier (multiplication n n) ( ( i -1 ) / 2 ) ) ;;
